% Created by: Syed Zain Masood
% Date: 30th April, 2009
% This code is intended to remove saturation effects from photographs

function X = SaturationRemoval(img,satMap)

[m n o] = size(img);
pix = m*n;
I = speye(pix);
Z = sparse([],[],[],pix,pix);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Cross-channel correlation calculated
alpha_rg = CrossChannel(img,1,2,3,satMap);
alpha_rb = CrossChannel(img,1,3,2,satMap);
alpha_gb = CrossChannel(img,2,3,1,satMap);

alpha_gr = 1./(alpha_rg+0.00001);
alpha_br = 1./(alpha_rb+0.00001);
alpha_bg = 1./(alpha_gb+0.00001);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Weights for following constarints calculated:
%   (i)   RED dependent on GREEN  (diagw1rg),
%   (ii)  RED dependent on BLUE   (diagw1rb),
%   (iii) GREEN dependent on RED  (diagw1gr),
%   (iv)  GREEN dependent on BLUE (diagw1gb),
%   (v)   BLUE dependent on RED   (diagw1rg),
%   (vi)  BLUE dependent on GREEN (diagw1rb),

w1rg = double(satMap(:,:,1) & (~satMap(:,:,2)));
w1rb = double(satMap(:,:,1) & (~satMap(:,:,3)));
w1gr = double(satMap(:,:,2) & (~satMap(:,:,1)));
w1gb = double(satMap(:,:,2) & (~satMap(:,:,3)));
w1br = double(satMap(:,:,3) & (~satMap(:,:,1)));
w1bg = double(satMap(:,:,3) & (~satMap(:,:,2)));

diagw1rg = spdiags(w1rg(:),0,length(w1rg(:)),length(w1rg(:)));
diagw1rb = spdiags(w1rb(:),0,length(w1rb(:)),length(w1rb(:)));
diagw1gr = spdiags(w1gr(:),0,length(w1gr(:)),length(w1gr(:)));
diagw1gb = spdiags(w1gb(:),0,length(w1gb(:)),length(w1gb(:)));
diagw1br = spdiags(w1br(:),0,length(w1br(:)),length(w1br(:)));
diagw1bg = spdiags(w1bg(:),0,length(w1bg(:)),length(w1bg(:)));

clear w1*

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Weights for smoothness constarints calculated.

satMapAll = satMap(2:end-1,2:end-1,:);
w2All = double(satMapAll(:,:,1) & satMapAll(:,:,2) & satMapAll(:,:,3));
diagw2A = spdiags(w2All(:),0,length(w2All(:)),length(w2All(:)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Reshaping color channels to diagonal matrices, respectively.
tempRed = img(:,:,1);
tempRed = reshape(tempRed,[],1);
diagRed = spdiags(tempRed,0,length(tempRed(:)),length(tempRed(:)));

tempGreen = img(:,:,2);
tempGreen = reshape(tempGreen,[],1);
diagGreen = spdiags(tempGreen,0,length(tempGreen(:)),length(tempGreen(:)));

tempBlue  = img(:,:,3);
tempBlue  = reshape(tempBlue,[],1);
diagBlue  = spdiags(tempBlue,0,length(tempBlue(:)),length(tempBlue(:)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Considering ratio of RED channel w.r.t GREEN and BLUE channels.
A1r = [I Z Z];
h1rg = diagGreen*alpha_rg;
h1rb = diagBlue*alpha_rb;

M = A1r'*diagw1rg*A1r  + A1r'*diagw1rb*A1r;
N = A1r'*diagw1rg*h1rg + A1r'*diagw1rb*h1rb;

% Considering ratio of GREEN channel w.r.t RED and BLUE channels.
A1g = [Z I Z];
h1gr = diagRed*alpha_gr;
h1gb = diagBlue*alpha_gb;

M = M + A1g'*diagw1gr*A1g  + A1g'*diagw1gb*A1g;
N = N + A1g'*diagw1gr*h1gr + A1g'*diagw1gb*h1gb;

% Considering ratio of BLUE channel w.r.t RED and GREEN channels.
A1b = [Z Z I];
h1br = diagRed*alpha_br;
h1bg = diagGreen*alpha_bg;

M = M + A1b'*diagw1br*A1b  + A1b'*diagw1bg*A1b;
N = N + A1b'*diagw1br*h1br + A1b'*diagw1bg*h1bg;

clear alpha_* diagRed diagGreen diagBlue h1* diagw1*

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% For Smoothness, neighboring pixels are taken into consideration
% Considering a 8-connected neighborhood for RED, GREEN and BLUE values.

diagw2 = diagw2A;

% Top Left Corner Pixel
A2 = filt2mat([-1 0  0; 0  1  0; 0  0  0],[m n]);
A2z = sparse([],[],[],size(A2,1),size(A2,2));
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
h2 = zeros(size(diagw2,1),1);
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Left Pixel
A2 = filt2mat([0 -1  0; 0  1  0; 0  0  0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Bottom Left Corner Pixel
A2 = filt2mat([0 0 -1; 0  1  0; 0  0  0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Top Pixel
A2 = filt2mat([0 0 0; -1  1  0; 0  0  0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Bottom Pixel
A2 = filt2mat([0 0 0;0 1 -1;0 0 0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Top Right Corner Pixel
A2 = filt2mat([0 0 0;0 1 0;-1 0 0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Right Pixel
A2 = filt2mat([0 0 0;0 1 0;0 -1 0],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;
clear A2 A2r A2g A2b

% Bottom Right Corner Pixel
A2 = filt2mat([0 0 0;0 1 0;0 0 -1],[m n]);
A2r = [A2 A2z A2z];
A2g = [A2z A2 A2z];
A2b = [A2z A2z A2];
M = M + A2r'*diagw2*A2r + A2g'*diagw2*A2g + A2b'*diagw2*A2b;
N = N + A2r'*diagw2*h2  + A2g'*diagw2*h2  + A2b'*diagw2*h2;

clear A2 A2r A2g A2b
clear diagw2* A2*

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If not saturated, RED, GREEN and BLUE pixels showed be close to observed.

h3r = tempRed;
h3g = tempGreen;
h3b = tempBlue;

w3r = double(~satMap(:,:,1));
w3g = double(~satMap(:,:,2));
w3b = double(~satMap(:,:,3));

diagw3r = spdiags(w3r(:),0,length(w3r(:)),length(w3r(:)));
diagw3g = spdiags(w3g(:),0,length(w3g(:)),length(w3g(:)));
diagw3b = spdiags(w3b(:),0,length(w3b(:)),length(w3b(:)));

M = M + A1r'*diagw3r*A1r + A1g'*diagw3g*A1g + A1b'*diagw3b*A1b;
N = N + A1r'*diagw3r*h3r + A1g'*diagw3g*h3g + A1b'*diagw3b*h3b;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Solving the system
X = M\N;
X = reshape(X,m,n,o);